/*
 * cdrom-cdaudio.c
 *
 *   Copyright (C) Ueno Katsuhiro 2000
 *
 * $Id: cdrom-cdaudio.c,v 1.1 2000/10/20 23:03:56 katsu Exp $
 */

#include <cdaudio.h>
#include <errno.h>
#include "ruby.h"
#include "rbcd.h"

#define unsupported(msg)   (errno = ENOSYS, rb_sys_fail(msg))

static int last_status = RBCD_NO_STATUS;



void rbcd_open(union rbcd_device *cdrom, const char *device)
{
	cdrom->fd = cd_init_device((char *)device);   /* shut up! */
	if (cdrom->fd < 0)
		rb_sys_fail(device);
}


void rbcd_close(union rbcd_device cdrom)
{
	cd_finish(cdrom.fd);
}


void rbcd_inject(union rbcd_device cdrom)
{
	if (cd_close(cdrom.fd) < 0)
		rb_sys_fail("cd_close");
}


void rbcd_eject(union rbcd_device cdrom)
{
	if (cd_eject(cdrom.fd) < 0)
		rb_sys_fail("cd_eject");
}


void rbcd_lock_door(union rbcd_device cdrom, int locked)
{
	unsupported(0);
}


void rbcd_get_track_range(union rbcd_device cdrom, int *start, int *end)
{
	struct disc_info disc;

	if (cd_stat(cdrom.fd, &disc) < 0)
		rb_sys_fail("cd_stat");
	if (!disc.disc_present)
		rb_raise(rb_eCDROMNoDisc, "disc is not present");
	if (disc.disc_mode == CDAUDIO_COMPLETED)
		last_status = RBCD_COMPLETED;

	*start = disc.disc_first_track;
	*end = disc.disc_total_tracks;
	/* *end = disc.disc_first_track + disc.disc_total_tracks - 1; */
}


void rbcd_get_track(union rbcd_device cdrom,
		    int track_no, struct rbcd_track *track)
{
	struct disc_info disc;

	if (cd_stat(cdrom.fd, &disc) < 0)
		rb_sys_fail("cd_stat");
	if (!disc.disc_present)
		rb_raise(rb_eCDROMNoDisc, "disc is not present");
	if (disc.disc_mode == CDAUDIO_COMPLETED)
		last_status = RBCD_COMPLETED;

	if (track_no < 0) {
		track->track_type = RBCD_LEADOUT;
		track->start_addr.lba = -1;
		track->start_addr.msf.min = disc.disc_length.minutes;
		track->start_addr.msf.sec = disc.disc_length.seconds;
		track->start_addr.msf.frame = disc.disc_length.frames;
	} else {
		if (disc.disc_track[track_no - 1].track_type
		    == CDAUDIO_TRACK_AUDIO)
			track->track_type = RBCD_AUDIO_TRACK;
		else
			track->track_type = RBCD_DATA_TRACK;
		track->start_addr.lba = -1;
		track->start_addr.msf.min =
			disc.disc_track[track_no - 1].track_pos.minutes;
		track->start_addr.msf.sec =
			disc.disc_track[track_no - 1].track_pos.seconds;
		track->start_addr.msf.frame =
			disc.disc_track[track_no - 1].track_pos.frames;
	}
}


int rbcd_get_last_session_addr(union rbcd_device cdrom,
			       struct rbcd_time *addr)
{
	return 0;
}


void rbcd_play_audio(union rbcd_device cdrom,
		     struct rbcd_time *start, struct rbcd_time *end)
{
	if (cd_play_frames(cdrom.fd,
			   cd_lba_to_frames(start->lba),
			   cd_lba_to_frames(end->lba)) < 0)
		rb_sys_fail("cd_play_frames");
}


void rbcd_pause_audio(union rbcd_device cdrom)
{
	if (cd_pause(cdrom.fd) < 0)
		rb_sys_fail("cd_pause");
}


void rbcd_resume_audio(union rbcd_device cdrom)
{
	if (cd_resume(cdrom.fd) < 0)
		rb_sys_fail("cd_resume");
}


void rbcd_stop_audio(union rbcd_device cdrom)
{
	if (cd_stop(cdrom.fd) < 0)
		rb_sys_fail("cd_stop");
}


void rbcd_read_audio(union rbcd_device cdrom,
		     struct rbcd_time *addr, int frames, char *dst)
{
	unsupported(0);
}

void rbcd_read_mode1(union rbcd_device cdrom,
		     struct rbcd_time *addr, int frames, char *dst)
{
	unsupported(0);
}

void rbcd_read_mode2(union rbcd_device cdrom,
		     struct rbcd_time *addr, int frames, char *dst)
{
	unsupported(0);
}

void rbcd_read_raw(union rbcd_device cdrom,
		   struct rbcd_time *addr, int frames, char *dst)
{
	unsupported(0);
}


void rbcd_get_status(union rbcd_device cdrom, struct rbcd_status *status)
{
	struct disc_status disc;

	if (cd_poll(cdrom.fd, &disc) < 0)
		rb_sys_fail("cd_poll");
	if (!disc.status_present)
		rb_raise(rb_eCDROMNoDisc, "disc is not present");

	switch(disc.status_mode) {
	  case CDAUDIO_PLAYING:
		status->status = RBCD_PLAYING;
		break;
	  case CDAUDIO_PAUSED:
		status->status = RBCD_PAUSED;
		break;
	  case CDAUDIO_COMPLETED:
		status->status = RBCD_COMPLETED;
		break;
	  default:
		status->status = last_status;
		break;
	}
	last_status = RBCD_NO_STATUS;
	status->track_no = disc.status_current_track;
	status->addr.lba = -1;
	status->addr.msf.min     = disc.status_disc_time.minutes;
	status->addr.msf.sec     = disc.status_disc_time.seconds;
	status->addr.msf.frame   = disc.status_disc_time.frames;
	status->track_time.min   = disc.status_track_time.minutes;
	status->track_time.sec   = disc.status_track_time.seconds;
	status->track_time.frame = disc.status_track_time.frames;
}


void rbcd_get_volume(union rbcd_device cdrom, int *left, int *right)
{
	struct disc_volume vol;

	if (cd_get_volume(cdrom.fd, &vol) < 0)
		rb_sys_fail("cd_get_volume");
	*left = vol.vol_front.left;
	*right = vol.vol_front.right;
}


void rbcd_set_volume(union rbcd_device cdrom, int left, int right)
{
	struct disc_volume vol;

	vol.vol_front.left = vol.vol_back.left = left;
	vol.vol_front.right = vol.vol_back.right = right;
	if (cd_set_volume(cdrom.fd, vol) < 0)
		rb_sys_fail("cd_get_volume");
}


int rbcd_num_slots(union rbcd_device cdrom)
{
	/* consider it unsupported */
	return 1;
}


int rbcd_change_slot(union rbcd_device cdrom, int slot)
{
	/* consider it unsupported */
	if (slot <= 0)
		return 0;
	unsupported(0);
}


int rbcd_is_ready(union rbcd_device cdrom, int slot)
{
	/* consider it unsupported */
	struct disc_status disc;

	if (slot > 0)
		unsupported(0);

	if (cd_poll(cdrom.fd, &disc) < 0)
		rb_sys_fail("cd_poll");
	if (disc.status_mode == CDAUDIO_COMPLETED)
		last_status = RBCD_COMPLETED;
	return (disc.status_present != 0);
}
