/*
 * cdrom-linux.c
 *
 *   Copyright (C) Ueno Katsuhiro 2000
 *
 * $Id: cdrom-linux.c,v 1.2 2000/10/20 22:55:12 katsu Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/cdrom.h>
#include "ruby.h"

#include "rbcd.h"


static void syserror(const char *msg) NORETURN;
static void syserror(const char *msg)
{
	if (errno == ENOMEDIUM)
		rb_raise(rb_eCDROMNoDisc, "%s - no disc found", msg);
	rb_sys_fail(msg);
}


#define rbcd_ioctl_1(dev,req)\
	if (ioctl((dev).fd,(req)) < 0) syserror(#req)
#define rbcd_ioctl(dev,req,p)\
	if (ioctl((dev).fd,(req),(p)) < 0) syserror(#req)



void rbcd_open(union rbcd_device *cdrom, const char *device)
{
	cdrom->fd = open(device, O_RDONLY | O_NONBLOCK);
	if (cdrom->fd < 0)
		syserror(device);
}



void rbcd_close(union rbcd_device cdrom)
{
	close(cdrom.fd);
}



void rbcd_inject(union rbcd_device cdrom)
{
	rbcd_ioctl_1(cdrom, CDROMCLOSETRAY);
}



void rbcd_eject(union rbcd_device cdrom)
{
	rbcd_ioctl_1(cdrom, CDROMEJECT);
}



void rbcd_lock_door(union rbcd_device cdrom, int locked)
{
	rbcd_ioctl(cdrom, CDROM_LOCKDOOR, locked);
}



void rbcd_get_track_range(union rbcd_device cdrom, int *start, int *end)
{
	struct cdrom_tochdr hdr;

	memset(&hdr, 0, sizeof(hdr));
	rbcd_ioctl(cdrom, CDROMREADTOCHDR, &hdr);
	*start = hdr.cdth_trk0;
	*end = hdr.cdth_trk1;
}



void rbcd_get_track(union rbcd_device cdrom,
		    int track_no, struct rbcd_track *track)
{
	struct cdrom_tocentry entry;

	memset(&entry, 0, sizeof(entry));
	entry.cdte_track = (track_no >= 0) ? track_no : CDROM_LEADOUT;
	entry.cdte_format = CDROM_LBA;
	rbcd_ioctl(cdrom, CDROMREADTOCENTRY, &entry);

	if (track_no < 0)
		track->track_type = RBCD_LEADOUT;
	else if ((entry.cdte_ctrl & CDROM_DATA_TRACK) == 0)
		track->track_type = RBCD_AUDIO_TRACK;
	else
		track->track_type = RBCD_DATA_TRACK;

	track->start_addr.lba = entry.cdte_addr.lba;
}


int rbcd_get_last_session_addr(union rbcd_device cdrom, struct rbcd_time *addr)
{
	struct cdrom_multisession ms;

	memset(&ms, 0, sizeof(ms));
	ms.addr_format = CDROM_LBA;
	if (ioctl(cdrom.fd, CDROMMULTISESSION, &ms) < 0) {
		if (errno == ENOSYS)
			return 0;
		else
			syserror("CDROMMULTISESSION");
	}
	if (!ms.xa_flag)
		return 0;
	addr->lba = ms.addr.lba;
	return 1;
}


void rbcd_play_audio(union rbcd_device cdrom,
		      struct rbcd_time *start, struct rbcd_time *end)
{
	struct cdrom_msf msf;

	memset(&msf, 0, sizeof(msf));
	msf.cdmsf_min0   = start->msf.min;
	msf.cdmsf_sec0   = start->msf.sec;
	msf.cdmsf_frame0 = start->msf.frame;
	msf.cdmsf_min1   = end->msf.min;
	msf.cdmsf_sec1   = end->msf.sec;
	msf.cdmsf_frame1 = end->msf.frame;
	rbcd_ioctl(cdrom, CDROMPLAYMSF, &msf);
}


void rbcd_pause_audio(union rbcd_device cdrom)
{
	rbcd_ioctl_1(cdrom, CDROMPAUSE);
}


void rbcd_resume_audio(union rbcd_device cdrom)
{
	rbcd_ioctl_1(cdrom, CDROMRESUME);
}


void rbcd_stop_audio(union rbcd_device cdrom)
{
	rbcd_ioctl_1(cdrom, CDROMSTOP);
}


void rbcd_read_audio(union rbcd_device cdrom,
		      struct rbcd_time *start, int frames, char *dst)
{
	struct cdrom_read_audio ra;

	memset(&ra, 0, sizeof(ra));
	ra.addr_format = CDROM_LBA;
	ra.addr.lba = start->lba;
	ra.nframes = frames;
	ra.buf = dst;
	rbcd_ioctl(cdrom, CDROMREADAUDIO, &ra);
}


static
void read__frame(union rbcd_device cdrom, int request, const char *reqstr,
		 int framesize, struct rbcd_time *start, int frames, char *dst)
{
	struct rbcd_time_msf addr = start->msf;
	struct cdrom_msf *msf;
	int n;

	for ( ; frames > 0; frames--) {
		msf = (struct cdrom_msf *)dst;
		memset(msf, 0, sizeof(struct cdrom_msf));
		msf->cdmsf_min0   = addr.min;
		msf->cdmsf_sec0   = addr.sec;
		msf->cdmsf_frame0 = addr.frame;
		n = ioctl(cdrom.fd, request, msf);
		if (n < 0)
			syserror(reqstr);
		dst += framesize;
		if (++addr.frame >= RBCD_MSF_FRAMES) {
			addr.frame = 0;
			if (++addr.sec >= 60)
				addr.sec = 0, addr.min++;
		}
	}
}

#define read_frame(cdrom,request,framesize,start,frames,dst)\
	read__frame(cdrom,request,#request,framesize,start,frames,dst)


void rbcd_read_mode1(union rbcd_device cdrom,
		     struct rbcd_time *start, int frames, char *dst)
{
	read_frame(cdrom, CDROMREADMODE1, RBCD_FRAMESIZE_MODE1,
		   start, frames, dst);
}


void rbcd_read_mode2(union rbcd_device cdrom,
		     struct rbcd_time *start, int frames, char *dst)
{
	read_frame(cdrom, CDROMREADMODE2, RBCD_FRAMESIZE_MODE2,
		   start, frames, dst);
}


void rbcd_read_raw(union rbcd_device cdrom,
		   struct rbcd_time *start, int frames, char *dst)
{
	read_frame(cdrom, CDROMREADRAW, RBCD_FRAMESIZE_RAW,
		   start, frames, dst);
}


void rbcd_get_status(union rbcd_device cdrom, struct rbcd_status *status)
{
	struct cdrom_subchnl sc;

	memset(&sc, 0, sizeof(sc));
	sc.cdsc_format = CDROM_MSF;
	rbcd_ioctl(cdrom, CDROMSUBCHNL, &sc);

	status->track_no = sc.cdsc_trk;
	status->addr.lba = -1;
	status->addr.msf.min     = sc.cdsc_absaddr.msf.minute;
	status->addr.msf.sec     = sc.cdsc_absaddr.msf.second;
	status->addr.msf.frame   = sc.cdsc_absaddr.msf.frame;
	status->track_time.min   = sc.cdsc_reladdr.msf.minute;
	status->track_time.sec   = sc.cdsc_reladdr.msf.second;
	status->track_time.frame = sc.cdsc_reladdr.msf.frame;

	switch (sc.cdsc_audiostatus) {
	  case CDROM_AUDIO_PLAY:
		status->status = RBCD_PLAYING;
		break;
	  case CDROM_AUDIO_PAUSED:
		status->status = RBCD_PAUSED;
		break;
	  case CDROM_AUDIO_COMPLETED:
		status->status = RBCD_COMPLETED;
		break;
	  default:
		status->status = RBCD_NO_STATUS;
	}
}


void rbcd_get_volume(union rbcd_device cdrom, int *left, int *right)
{
	struct cdrom_volctrl vol;

	memset(&vol, 0, sizeof(vol));
	rbcd_ioctl(cdrom, CDROMVOLREAD, &vol);
	*left = vol.channel0;
	*right = vol.channel1;
}


void rbcd_set_volume(union rbcd_device cdrom, int left, int right)
{
	struct cdrom_volctrl vol;

	memset(&vol, 0, sizeof(vol));
	vol.channel0 = left;
	vol.channel1 = right;
	vol.channel2 = 0;
	vol.channel3 = 0;
	rbcd_ioctl(cdrom, CDROMVOLCTRL, &vol);
}


int rbcd_num_slots(union rbcd_device cdrom)
{
	int n;

	n = ioctl(cdrom.fd, CDROM_CHANGER_NSLOTS);
	if (n < 0) {
		if (errno == ENOSYS)
			n = 1;
		else
			syserror("CDROM_CHANGER_NSLOTS");
	}
	return n;
}


int rbcd_change_slot(union rbcd_device cdrom, int slot)
{
	int n;

	if (slot < 0)
		slot = CDSL_CURRENT;
	n = ioctl(cdrom.fd, CDROM_SELECT_DISC, slot);
	if (n < 0) {
		if (errno == ENOSYS && (slot == CDSL_CURRENT || slot == 0))
			n = 0;
		else
			syserror("CDROM_SELECT_DISC");
	}
	return n;
}


int rbcd_is_ready(union rbcd_device cdrom, int slot)
{
	int n;

	if (slot < 0)
		slot = CDSL_CURRENT;
	n = ioctl(cdrom.fd, CDROM_DRIVE_STATUS, slot);
	if (n < 0)
		syserror("CDROM_SELECT_DISC");
	return (n == CDS_DISC_OK);
}
