/*
 * cdrom.c
 *
 *   Copyright (C) Ueno Katsuhiro 2000
 *
 * $Id: cdrom.c,v 1.1.1.1 2000/10/15 06:13:34 katsu Exp $
 */

#include "ruby.h"
#include "rbcd.h"

#if defined(__GNUC__) && __GNUC__ >= 2 && !defined(RUBY_NO_INLINE)
#define INLINE  __inline__
#else
#define INLINE
#endif


#ifndef DEFAULT_DEVICE
#define DEFAULT_DEVICE  "/dev/cdrom"
#endif

#define CDEXTRA_GAP  11400    /* sectors between two sessions */


static VALUE cTime, cTrack, cStatus;
VALUE rb_eCDROMNoDisc;


/* for debug */
#define WARN_ADDR(prefix,a)  rb_warn(prefix "%d [%02d:%02d.%02d]",\
				     a.lba,a.msf.min,a.msf.sec,a.msf.frame)


#define LBA_F(lba)  ((lba) % RBCD_MSF_FRAMES)
#define LBA_S(lba)  (((lba) / RBCD_MSF_FRAMES) % 60)
#define LBA_M(lba)  (((lba) / RBCD_MSF_FRAMES) / 60)

#define MSF_ADD(msf, n) do {					\
	(msf).frame += LBA_F(n);				\
	if (LBA_F(n) && (msf).frame >= RBCD_MSF_FRAMES)		\
		(msf).sec++, (msf).frame -= RBCD_MSF_FRAMES;	\
	(msf).sec += LBA_S(n);					\
	if ((msf).sec >= 60)					\
		(msf).min++, (msf).sec -= 60;			\
	(msf).min += LBA_M(n);					\
} while (0)

#define MSF_SUB(msf, n) do {					\
	(msf).frame -= LBA_F(n);				\
	if (LBA_F(n) && (msf).frame < 0)			\
		(msf).sec--, (msf).frame += RBCD_MSF_FRAMES;	\
	(msf).sec -= LBA_S(n);					\
	if ((msf).sec < 0)					\
		(msf).min--, (msf).sec += 60;			\
	(msf).min -= LBA_M(n);					\
} while (0)


#define LBA_CHECK(lba) do {						\
	if ((lba) < 0)							\
		rb_raise(rb_eArgError, "time must be positive");	\
	if (!POSFIXABLE(lba))						\
		rb_raise(rb_eArgError, "time out of range");		\
} while (0)


static
void lba_to_msf(int lba, struct rbcd_time_msf *msf)
{
	msf->frame = lba % RBCD_MSF_FRAMES;
	msf->sec = (lba / RBCD_MSF_FRAMES) % 60;
	msf->min = (lba / RBCD_MSF_FRAMES) / 60;
}


#define MSF_TO_LBA(msf)\
	((msf)->min * 60 + (msf)->sec) * RBCD_MSF_FRAMES + (msf)->frame


static
void msf_normalize(struct rbcd_time_msf *msf)
{
	if (msf->frame >= 75) {
		msf->sec += msf->frame / 75;
		msf->frame %= 75;
	}
	else if (msf->frame < 0) {
		msf->sec += (msf->frame + 1) / 75 - 1;
		msf->frame = (msf->frame + 1) % 75 - 1 + 75;
	}
	if (msf->sec >= 60) {
		msf->min += msf->sec / 60;
		msf->sec %= 60;
	}
	else if (msf->sec < 0) {
		msf->min += (msf->sec + 1) / 60 - 1;
		msf->sec = (msf->sec + 1) % 60 - 1 + 60;
	}
}




/*
  rbcd_time --- (msf - MSF_OFFSET) --> cdtime
  rbcd_time <-- (msf + MSF_OFFSET) --- cdtime
*/

struct cdtime {
	int lba;
	struct rbcd_time_msf msf;   /* MSF without offset */
};



static
void value_to_addr(VALUE obj, struct rbcd_time *addr)
{
	if (rb_obj_is_kind_of(obj, cTime)) {
		struct cdtime *time = (struct cdtime *)DATA_PTR(obj);
		addr->lba = time->lba;
		addr->msf = time->msf;
		MSF_ADD(addr->msf, RBCD_MSF_OFFSET);
	} else {
		addr->lba = NUM2INT(obj);
		LBA_CHECK(addr->lba);
		lba_to_msf(addr->lba + RBCD_MSF_OFFSET, &addr->msf);
	}
}



static
void addr_complete(struct rbcd_time *addr)
{
	if (addr->lba >= 0) {
		if (!POSFIXABLE(addr->lba))
			rb_bug("lba out of range: %d", addr->lba);
		lba_to_msf(addr->lba + RBCD_MSF_OFFSET, &addr->msf);
	} else {
		msf_normalize(&addr->msf);
		if (addr->msf.min < 0)
			rb_bug("negative msf: %02d:%02d.%02d",
			       addr->msf.min, addr->msf.sec, addr->msf.frame);
		addr->lba = MSF_TO_LBA(&addr->msf) - RBCD_MSF_OFFSET;
#if 0
		if (addr->lba < 0)
			rb_warning("negative lba: %d", addr->lba);
#endif
	}
}



static
VALUE cdtime_new(const struct rbcd_time *addr)
{
	VALUE obj;
	struct cdtime *time;

	obj = Data_Make_Struct(cTime, struct cdtime, 0, 0, time);

	/* addr_complete(addr); */
	if (addr->lba >= 0) {
		time->lba = addr->lba;
		time->msf = addr->msf;
		MSF_SUB(time->msf, RBCD_MSF_OFFSET);
	} else {
		time->lba = 0;
		time->msf.min = time->msf.sec = time->msf.frame = 0;
	}
	return obj;
}



static
VALUE cdtime_new_msf(struct rbcd_time_msf *msf)
{
	VALUE obj;
	struct cdtime *time;

	obj = Data_Make_Struct(cTime, struct cdtime, 0, 0, time);
	time->lba = MSF_TO_LBA(msf);
	time->msf = *msf;
	return obj;
}



static
VALUE cdtime_new_obj(VALUE klass, int lba)
{
	VALUE obj;
	struct cdtime *time;

	LBA_CHECK(lba);
	obj = Data_Make_Struct(cTime, struct cdtime, 0, 0, time);
	time->lba = lba;
	lba_to_msf(lba, &time->msf);
	return obj;
}


static
int value_to_lba(VALUE obj)
{
	if (rb_obj_is_kind_of(obj, cTime))
		return ((struct cdtime *)DATA_PTR(obj))->lba;
	else {
		int lba = NUM2INT(obj);
		LBA_CHECK(lba);
		return lba;
	}
}


VALUE rb_cdtime_s_new(VALUE klass, VALUE lba)
{
	return cdtime_new_obj(klass, value_to_lba(lba));
}


VALUE rb_cdtime_s_at(VALUE klass, VALUE sec)
{
	int lba;
	if (FIXNUM_P(sec))
		lba = FIX2INT(sec) * 75;
	else
		lba = NUM2DBL(sec) * 75;
	return cdtime_new_obj(klass, lba);
}



static INLINE
struct cdtime *get_cdtime(VALUE obj)
{
	struct cdtime *t;
	Data_Get_Struct(obj, struct cdtime, t);
	return t;
}

VALUE rb_cdtime_to_a(VALUE obj)
{
	struct cdtime *t = get_cdtime(obj);
	return rb_ary_new3(4, INT2FIX(t->lba), INT2FIX(t->msf.min),
			   INT2FIX(t->msf.sec), INT2FIX(t->msf.frame));
}

VALUE rb_cdtime_seconds(VALUE obj)
{
	return rb_float_new((double)(get_cdtime(obj)->lba) / RBCD_MSF_FRAMES);
}

VALUE rb_cdtime_to_i(VALUE obj)
{
	return INT2FIX(get_cdtime(obj)->lba);
}

VALUE rb_cdtime_min(VALUE obj)
{
	return INT2FIX(get_cdtime(obj)->msf.min);
}

VALUE rb_cdtime_sec(VALUE obj)
{
	return INT2FIX(get_cdtime(obj)->msf.sec);
}

VALUE rb_cdtime_frame(VALUE obj)
{
	return INT2FIX(get_cdtime(obj)->msf.frame);
}

VALUE rb_cdtime_plus(VALUE obj, VALUE other)
{
	return cdtime_new_obj(RBASIC(obj)->klass,
			      get_cdtime(obj)->lba + value_to_lba(other));
}

VALUE rb_cdtime_minus(VALUE obj, VALUE other)
{
	return cdtime_new_obj(RBASIC(obj)->klass,
			      get_cdtime(obj)->lba - value_to_lba(other));
}

VALUE rb_cdtime_cmp(VALUE obj, VALUE other)
{
	int lba1 = get_cdtime(obj)->lba;
	int lba2 = value_to_lba(other);

	if (lba1 > lba2)
		return INT2FIX(1);
	else if (lba1 == lba2)
		return INT2FIX(0);
	else
		return INT2FIX(-1);
}

VALUE rb_cdtime_to_real(VALUE obj, VALUE other)
{
	return cdtime_new_obj(RBASIC(obj)->klass,
			      get_cdtime(obj)->lba + RBCD_MSF_OFFSET);
}

VALUE rb_cdtime_inspect(VALUE obj)
{
	struct cdtime *t = get_cdtime(obj);
	VALUE argv[5];

	argv[0] = rb_str_new2("%d [%02d:%02d.%02d]");
	argv[1] = INT2FIX(t->lba);
	argv[2] = INT2FIX(t->msf.min);
	argv[3] = INT2FIX(t->msf.sec);
	argv[4] = INT2FIX(t->msf.frame);
	return rb_f_sprintf(5, argv);
}




struct track {
	struct rbcd_track t;
	VALUE start_addr;
};


static
void track_mark(struct track *track)
{
	rb_gc_mark(track->start_addr);
}


static
VALUE track_new(struct rbcd_track *t)
{
	struct track *track;
	VALUE obj;

	obj = Data_Make_Struct(cTrack, struct track, track_mark, 0, track);
	track->t = *t;
	track->start_addr = Qnil;
	return obj;
}



static INLINE
struct track *get_track(VALUE obj)
{
	struct track *track;
	Data_Get_Struct(obj, struct track, track);
	return track;
}

VALUE rb_track_no(VALUE obj)
{
	return INT2FIX(get_track(obj)->t.track_no);
}

VALUE rb_track_start_addr(VALUE obj)
{
	struct track *track = get_track(obj);
	if (NIL_P(track->start_addr))
		track->start_addr = cdtime_new(&track->t.start_addr);
	return track->start_addr;
}

VALUE rb_track_type(VALUE obj)
{
	static const char * const desc[] = { RBCD_TRACK_TYPE_STR };
	return rb_str_new2(desc[get_track(obj)->t.track_type]);
}

VALUE rb_track_audio_p(VALUE obj)
{
	return get_track(obj)->t.track_type == RBCD_AUDIO_TRACK ? Qtrue:Qfalse;
}

VALUE rb_track_data_p(VALUE obj)
{
	return get_track(obj)->t.track_type == RBCD_DATA_TRACK ? Qtrue:Qfalse;
}

VALUE rb_track_leadout_p(VALUE obj)
{
	return get_track(obj)->t.track_type == RBCD_LEADOUT ? Qtrue : Qfalse;
}

VALUE rb_track_inspect(VALUE obj)
{
	struct track *track = get_track(obj);
	VALUE argv[9];

	argv[0] = rb_str_new2("#<%s:0x%x track_no=%d, "
			      "start_addr=%d@%02d:%02d.%02d, type=%s>");
	argv[1] = rb_str_new2(rb_class2name(CLASS_OF(obj)));
	argv[2] = rb_obj_id(obj);
	argv[3] = INT2FIX(track->t.track_no);
	argv[4] = INT2FIX(track->t.start_addr.lba);
	argv[5] = INT2FIX(track->t.start_addr.msf.min);
	argv[6] = INT2FIX(track->t.start_addr.msf.sec);
	argv[7] = INT2FIX(track->t.start_addr.msf.frame);
	argv[8] = rb_track_type(obj);
	return rb_f_sprintf(9, argv);
}




struct status {
	struct rbcd_status s;
	VALUE addr, track_time;
};


static
void status_mark(struct status *status)
{
	rb_gc_mark(status->addr);
	rb_gc_mark(status->track_time);
}


static
VALUE status_new(struct rbcd_status *s)
{
	struct status *status;
	VALUE obj;

	obj = Data_Make_Struct(cStatus, struct status, status_mark, 0, status);
	status->s = *s;
	status->addr = status->track_time = Qnil;
	return obj;
}



static INLINE
struct status *get_status(VALUE obj)
{
	struct status *status;
	Data_Get_Struct(obj, struct status, status);
	return status;
}

VALUE rb_status_status(VALUE obj)
{
	static const char * const desc[] = { RBCD_STATUS_STR };
	return rb_str_new2(desc[get_status(obj)->s.status]);
}

VALUE rb_status_track_no(VALUE obj)
{
	return INT2FIX(get_status(obj)->s.track_no);
}

VALUE rb_status_addr(VALUE obj)
{
	struct status *status = get_status(obj);
	if (NIL_P(status->addr))
		status->addr = cdtime_new(&status->s.addr);
	return status->addr;
}

VALUE rb_status_track_time(VALUE obj)
{
	struct status *status = get_status(obj);
	if (NIL_P(status->track_time))
		status->track_time = cdtime_new_msf(&status->s.track_time);
	return status->track_time;
}

VALUE rb_status_playing_p(VALUE obj)
{
	return get_status(obj)->s.status == RBCD_PLAYING ? Qtrue : Qfalse;
}

VALUE rb_status_paused_p(VALUE obj)
{
	return get_status(obj)->s.status == RBCD_PAUSED ? Qtrue : Qfalse;
}

VALUE rb_status_completed_p(VALUE obj)
{
	return get_status(obj)->s.status == RBCD_COMPLETED ? Qtrue : Qfalse;
}

VALUE rb_status_no_status_p(VALUE obj)
{
	return get_status(obj)->s.status == RBCD_NO_STATUS ? Qtrue : Qfalse;
}

VALUE rb_status_inspect(VALUE obj)
{
	struct status *status = get_status(obj);
	VALUE argv[12];

	argv[0] = rb_str_new2("#<%s:0x%x track_no=%d, "
			      "addr=%d@%02d:%02d.%02d, "
			      "track_time=%02d:%02d.%02d, "
			      "status=%s>");
	argv[1] = rb_str_new2(rb_class2name(CLASS_OF(obj)));
	argv[2] = rb_obj_id(obj);
	argv[3] = INT2FIX(status->s.track_no);
	argv[4] = INT2FIX(status->s.addr.lba);
	argv[5] = INT2FIX(status->s.addr.msf.min);
	argv[6] = INT2FIX(status->s.addr.msf.sec);
	argv[7] = INT2FIX(status->s.addr.msf.frame);
	argv[8] = INT2FIX(status->s.track_time.min);
	argv[9] = INT2FIX(status->s.track_time.sec);
	argv[10] = INT2FIX(status->s.track_time.frame);
	argv[11] = rb_status_status(obj);
	return rb_f_sprintf(12, argv);
}




struct cdrom {
	VALUE closed;
	VALUE device_name;
	union rbcd_device device;
};


static
void cdrom_mark(struct cdrom *cdrom)
{
	rb_gc_mark(cdrom->device_name);
}


static
void cdrom_free(struct cdrom *cdrom)
{
	rbcd_close(cdrom->device);
	free(cdrom);
}


static
struct cdrom *get_cdrom(VALUE obj)
{
	struct cdrom *cdrom;

	Data_Get_Struct(obj, struct cdrom, cdrom);
	if (cdrom->closed != Qfalse)
		rb_raise(rb_eIOError, "closed CD-ROM device");
	return cdrom;
}



VALUE rb_cdrom_s_open(VALUE klass, VALUE device)
{
	struct cdrom *cdrom;
	VALUE obj;

	rb_secure(4);
	obj = Data_Make_Struct(klass, struct cdrom, cdrom_mark, cdrom_free,
			       cdrom);
	cdrom->closed = Qfalse;
	cdrom->device_name = rb_str_to_str(device);
	rb_str_freeze(cdrom->device_name);
	rbcd_open(&cdrom->device, RSTRING(cdrom->device_name)->ptr);
	return obj;
}



VALUE rb_cdrom_close(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	rbcd_close(cdrom->device);
	return Qnil;
}



VALUE rb_cdrom_closed_p(VALUE obj)
{
	struct cdrom *cdrom;
	Data_Get_Struct(obj, struct cdrom, cdrom);
	return cdrom->closed;
}



VALUE rb_cdrom_device_name(VALUE obj)
{
	return get_cdrom(obj)->device_name;
}



VALUE rb_cdrom_inject(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	rb_secure(2);
	rbcd_inject(cdrom->device);
	return Qnil;
}



VALUE rb_cdrom_eject(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	rb_secure(2);
	rbcd_eject(cdrom->device);
	return Qnil;
}



VALUE rb_cdrom_unlock_door(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	rb_secure(2);
	rbcd_lock_door(cdrom->device, 0);
	return Qnil;
}



VALUE rb_cdrom_lock_door(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	rb_secure(2);

	rbcd_lock_door(cdrom->device, 1);
	if (rb_block_given_p())
		return rb_ensure(rb_yield, obj, rb_cdrom_unlock_door, obj);
	return Qnil;
}



VALUE rb_cdrom_get_tracks(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	struct rbcd_track track, extra;
	int cdextra, n, last;
	VALUE dst;

	rb_secure(2);
	cdextra = rbcd_get_last_session_addr(cdrom->device, &extra.start_addr);
	if (cdextra) {
		addr_complete(&extra.start_addr);
		if (extra.start_addr.lba < CDEXTRA_GAP) {
			rb_warn("last session lba too small: %d",
				extra.start_addr.lba);
			cdextra = 0;
		}
	}
	rbcd_get_track_range(cdrom->device, &n, &last);
	dst = rb_ary_new2(last - n + 1 + 1 + cdextra);

	for (; n <= last; n++) {
		track.track_no = n;
		rbcd_get_track(cdrom->device, n, &track);
		addr_complete(&track.start_addr);
		if (cdextra && track.start_addr.lba == extra.start_addr.lba) {
			extra.track_no = -1;
			extra.track_type = RBCD_LEADOUT;
			extra.start_addr.lba -= CDEXTRA_GAP;
			MSF_SUB(extra.start_addr.msf, CDEXTRA_GAP);
			rb_ary_push(dst, track_new(&extra));
			cdextra = 0;
		}
		rb_ary_push(dst, track_new(&track));
	}
	rbcd_get_track(cdrom->device, -1, &track);
	addr_complete(&track.start_addr);
	track.track_no = -1;
	track.track_type = RBCD_LEADOUT;
	rb_ary_push(dst, track_new(&track));
	return dst;
}



VALUE rb_cdrom_size(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	struct rbcd_track track;

	rb_secure(2);
	rbcd_get_track(cdrom->device, -1, &track);
	addr_complete(&track.start_addr);
	return cdtime_new(&track.start_addr);
}



VALUE rb_cdrom_play_audio(int argc, VALUE *argv, VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	VALUE v_start, v_end;
	struct rbcd_track track;
	struct rbcd_time start;
	struct rbcd_time * const end = &track.start_addr;

	rb_secure(2);
	rb_scan_args(argc, argv, "02", &v_start, &v_end);
	if (!NIL_P(v_start))
		value_to_addr(v_start, &start);
	else {
		start.lba = 0;
		start.msf.min   = RBCD_MSF_OFFSET % RBCD_MSF_FRAMES;
		start.msf.sec   = (RBCD_MSF_OFFSET / RBCD_MSF_FRAMES) % 60;
		start.msf.frame = (RBCD_MSF_OFFSET / RBCD_MSF_FRAMES) / 60;
	}
	if (!NIL_P(v_end))
		value_to_addr(v_end, end);
	else {
		rbcd_get_track(cdrom->device, -1, &track);
		addr_complete(end);
	}
	rbcd_play_audio(cdrom->device, &start, end);
	return Qnil;
}



VALUE rb_cdrom_pause_audio(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	rb_secure(2);
	rbcd_pause_audio(cdrom->device);
	return Qnil;
}



VALUE rb_cdrom_resume_audio(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	rb_secure(2);
	rbcd_resume_audio(cdrom->device);
	return Qnil;
}



VALUE rb_cdrom_stop_audio(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	rb_secure(2);
	rbcd_stop_audio(cdrom->device);
	return Qnil;
}



static INLINE
VALUE cdrom_read(int argc, VALUE *argv, union rbcd_device dev, int brocksize,
		 void (*read)(union rbcd_device,struct rbcd_time*,int,char*))
{
	VALUE dst, v_addr, v_frames;
	struct rbcd_time addr;
	int frames;

	rb_secure(2);
	rb_scan_args(argc, argv, "11", &v_addr, &v_frames);
	value_to_addr(v_addr, &addr);
	frames = (argc == 1) ? 1 : NUM2INT(v_frames);
	if (frames < 0)
		rb_raise(rb_eArgError, "negative length %d given", frames);
	dst = rb_tainted_str_new(0, frames * brocksize * 2);
	read(dev, &addr, frames, RSTRING(dst)->ptr);
	return dst;
}



VALUE rb_cdrom_read_audio(int argc, VALUE *argv, VALUE obj)
{
	return cdrom_read(argc, argv, get_cdrom(obj)->device,
			  RBCD_FRAMESIZE_AUDIO, rbcd_read_audio);
}


VALUE rb_cdrom_read_mode1(int argc, VALUE *argv, VALUE obj)
{
	return cdrom_read(argc, argv, get_cdrom(obj)->device,
			  RBCD_FRAMESIZE_MODE1, rbcd_read_mode1);
}


VALUE rb_cdrom_read_mode2(int argc, VALUE *argv, VALUE obj)
{
	return cdrom_read(argc, argv, get_cdrom(obj)->device,
			  RBCD_FRAMESIZE_MODE2, rbcd_read_mode2);
}


VALUE rb_cdrom_read_raw(int argc, VALUE *argv, VALUE obj)
{
	return cdrom_read(argc, argv, get_cdrom(obj)->device,
			  RBCD_FRAMESIZE_RAW, rbcd_read_raw);
}



VALUE rb_cdrom_get_status(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	struct rbcd_status status;

	rb_secure(2);
	rbcd_get_status(cdrom->device, &status);
	addr_complete(&status.addr);
	msf_normalize(&status.track_time);
	return status_new(&status);
}



VALUE rb_cdrom_get_volume(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	int left, right;

	rb_secure(2);
	rbcd_get_volume(cdrom->device, &left, &right);
	if (left < 0 || left > 255 || right < 0 || right > 255)
		rb_bug("volume out of range: (%d, %d)", left, right);
	return rb_assoc_new(INT2FIX(left), INT2FIX(right));
}



VALUE rb_cdrom_set_volume(int argc, VALUE *argv, VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	VALUE l, r;
	int left, right;

	rb_secure(2);
	rb_scan_args(argc, argv, "11", &l, &r);
	left = NUM2INT(l);
	right = NIL_P(r) ? left : NUM2INT(r);
	if (left < 0 || left > 255 || right < 0 || right > 255)
		rb_raise(rb_eArgError, "volume out of range");
	rbcd_set_volume(cdrom->device, left, right);
	return Qnil;
}



VALUE rb_cdrom_num_slots(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	int n;

	rb_secure(2);
	n = rbcd_num_slots(cdrom->device);
	if (n < 1 || !POSFIXABLE(n))
		rb_bug("num_slots out of range: %d", n);
	return INT2FIX(n);
}



VALUE rb_cdrom_change_slot(VALUE obj, VALUE arg)
{
	struct cdrom *cdrom = get_cdrom(obj);
	int n;

	rb_secure(2);
	if (NIL_P(arg))
		n = -1;
	else {
		n = NUM2INT(arg);
		if (n < 0)
			rb_raise(rb_eArgError, "slot number must be positive");
	}

	n = rbcd_change_slot(cdrom->device, n);
	if (n < 0 || !POSFIXABLE(n))
		rb_bug("num_slots out of range: %d", n);
	return INT2FIX(n);
}



VALUE rb_cdrom_current_slot(VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	int n;

	rb_secure(2);
	n = rbcd_change_slot(cdrom->device, -1);
	if (n < 0 || !POSFIXABLE(n))
		rb_bug("num_slots out of range: %d", n);
	return INT2FIX(n);
}



VALUE rb_cdrom_ready_p(int argc, VALUE *argv, VALUE obj)
{
	struct cdrom *cdrom = get_cdrom(obj);
	VALUE slot;
	int n;

	rb_secure(2);
	rb_scan_args(argc, argv, "01", &slot);
	if (NIL_P(slot))
		n = -1;
	else {
		Check_Type(slot, T_FIXNUM);
		n = FIX2INT(slot);
	}
	return rbcd_is_ready(cdrom->device, n) ? Qtrue : Qfalse;
}




void Init_cdrom(void)
{
	VALUE cCDROM, cError;

	cCDROM = rb_define_class("CDROM", rb_cObject);
	cError = rb_define_class_under(cCDROM, "Error", rb_eStandardError);
	rb_eCDROMNoDisc = rb_define_class_under(cCDROM, "NoDisc", cError);

	rb_define_singleton_method(cCDROM, "new", rb_cdrom_s_open, 1);
	rb_define_singleton_method(cCDROM, "open", rb_cdrom_s_open, 1);
	rb_define_method(cCDROM, "close", rb_cdrom_close, 0);
	rb_define_method(cCDROM, "closed?", rb_cdrom_closed_p, 0);
	rb_define_method(cCDROM, "path", rb_cdrom_device_name, 0);
	rb_define_method(cCDROM, "device", rb_cdrom_device_name, 0);
	rb_define_method(cCDROM, "inject", rb_cdrom_inject, 0);
	rb_define_method(cCDROM, "eject", rb_cdrom_eject, 0);
	rb_define_method(cCDROM, "unlock_door", rb_cdrom_unlock_door, 0);
	rb_define_method(cCDROM, "lock_door", rb_cdrom_lock_door, 0);
	rb_define_method(cCDROM, "get_tracks", rb_cdrom_get_tracks, 0);
	rb_define_method(cCDROM, "size", rb_cdrom_size, 0);
	rb_define_method(cCDROM, "length", rb_cdrom_size, 0);
	rb_define_method(cCDROM, "play_audio", rb_cdrom_play_audio, -1);
	rb_define_method(cCDROM, "pause_audio", rb_cdrom_pause_audio, 0);
	rb_define_method(cCDROM, "resume_audio", rb_cdrom_resume_audio, 0);
	rb_define_method(cCDROM, "stop_audio", rb_cdrom_stop_audio, 0);
	rb_define_method(cCDROM, "read_audio", rb_cdrom_read_audio, -1);
	rb_define_method(cCDROM, "read_sector", rb_cdrom_read_mode1, -1);
	rb_define_method(cCDROM, "read_mode1", rb_cdrom_read_mode1, -1);
	rb_define_method(cCDROM, "read_mode2", rb_cdrom_read_mode2, -1);
	rb_define_method(cCDROM, "read_raw", rb_cdrom_read_raw, -1);
	rb_define_method(cCDROM, "get_status", rb_cdrom_get_status, 0);
	rb_define_method(cCDROM, "get_volume", rb_cdrom_get_volume, 0);
	rb_define_method(cCDROM, "set_volume", rb_cdrom_set_volume, -1);
	rb_define_method(cCDROM, "num_slots", rb_cdrom_num_slots, 0);
	rb_define_method(cCDROM, "change_slot", rb_cdrom_change_slot, 1);
	rb_define_method(cCDROM, "current_slot", rb_cdrom_current_slot, 0);
	rb_define_method(cCDROM, "ready?", rb_cdrom_ready_p, -1);
	rb_define_const(cCDROM, "DefaultDevice", rb_str_new2(DEFAULT_DEVICE));

	cTime = rb_define_class_under(cCDROM, "Time", rb_cObject);
	rb_include_module(cTime, rb_mComparable);
	rb_define_singleton_method(cTime, "new", rb_cdtime_s_new, 1);
	rb_define_singleton_method(cTime, "at", rb_cdtime_s_at, 1);
	rb_define_method(cTime, "to_a", rb_cdtime_to_a, 0);
	rb_define_method(cTime, "seconds", rb_cdtime_seconds, 0);
	rb_define_method(cTime, "to_i", rb_cdtime_to_i, 0);
	rb_define_method(cTime, "min", rb_cdtime_min, 0);
	rb_define_method(cTime, "sec", rb_cdtime_sec, 0);
	rb_define_method(cTime, "frame", rb_cdtime_frame, 0);
	rb_define_method(cTime, "+", rb_cdtime_plus, 1);
	rb_define_method(cTime, "-", rb_cdtime_minus, 1);
	rb_define_method(cTime, "<=>", rb_cdtime_cmp, 1);
	rb_define_method(cTime, "real", rb_cdtime_to_real, 0);
	rb_define_method(cTime, "inspect", rb_cdtime_inspect, 0);

	cTrack = rb_define_class_under(cCDROM, "Track", rb_cObject);
	rb_define_method(cTrack, "track_no", rb_track_no, 0);
	rb_define_method(cTrack, "start_addr", rb_track_start_addr, 0);
	rb_define_method(cTrack, "addr", rb_track_start_addr, 0);
	rb_define_method(cTrack, "track_type", rb_track_type, 0);
	rb_define_method(cTrack, "audio_track?", rb_track_audio_p, 0);
	rb_define_method(cTrack, "data_track?", rb_track_data_p, 0);
	rb_define_method(cTrack, "leadout?", rb_track_leadout_p, 0);
	rb_define_method(cTrack, "inspect", rb_track_inspect, 0);

	cStatus = rb_define_class_under(cCDROM, "Status", rb_cObject);
	rb_define_method(cStatus, "status", rb_status_status, 0);
	rb_define_method(cStatus, "track_no", rb_status_track_no, 0);
	rb_define_method(cStatus, "addr", rb_status_addr, 0);
	rb_define_method(cStatus, "track_time", rb_status_track_time, 0);
	rb_define_method(cStatus, "playing?", rb_status_playing_p, 0);
	rb_define_method(cStatus, "paused?", rb_status_paused_p, 0);
	rb_define_method(cStatus, "completed?", rb_status_completed_p, 0);
	rb_define_method(cStatus, "no_status?", rb_status_no_status_p, 0);
	rb_define_method(cStatus, "inspect", rb_status_inspect, 0);
}
