#
# rbcd-tty.rb
#
#   Copyright (C) Ueno Katsuhiro 2000
#
# $Id: rbcd-tty.rb,v 1.1.1.1 2000/10/15 06:13:33 katsu Exp $
#

require 'rbcd'


class Console

  def initialize(input = STDIN, out = STDOUT)
    @input = input
    @out = out
    @stroke = ''
    @prompt = ''
    @ready = false
    @on_idle = nil
    @abbrev = []
  end


  def on_idle(&proc)
    @on_idle = proc
  end

  def add_word(*words)
    @abbrev.concat words.flatten
    @abbrev.sort!
    @abbrev.uniq!
  end


  def print_prompt
    @out.print "\r", @prompt, '> ', @stroke
  end
  private :print_prompt


  def prompt=(str)
    if @ready then
      n = @prompt.size + @stroke.size - str.size
      @out.print("\b" * n + ' ' * n) if n > 0
    end
    @prompt = str
    print_prompt if @ready
  end


  def write(str)
    @out.print "\r\n" if @ready
    @out.print str.gsub(/\r?\n/, "\r\n")
    @out.print "\r\n" unless str[-1] == ?\n
    print_prompt if @ready
  end

  def print(*args)
    write args.join
  end

  def printf(format, *args)
    write sprintf(format, *args)
  end


  def do_stty(arg)
    if @input.tty? then
      fork {
        STDIN.reopen @input unless @input == STDIN
        exec 'stty ' + arg
      }
      Process.wait
    end
  end
  private :do_stty


  def start
    catch(:close) {
      @out.sync = true
      do_stty 'raw -echo'
      begin
        @ready = true
        print_prompt
        loop {
          @on_idle.call if @on_idle
          next unless IO.select([@input],[],[],0.1)
          case (c = @input.getc)
          when ?\n, ?\r then
            @out.print "\r\n"
            @ready = false
            yield @stroke
            @ready = true
            @stroke = ''
            print_prompt
          when ?\t then
            n = @stroke.size
            words = @abbrev.select { |w| w[0, n] == @stroke }
            if words.empty? then
              next
            elsif words.size == 1 then
              s = (words[0][n, words[0].size] or '') + ' '
              @out.print s
              @stroke << s
            else
              i = 0
              (words[0].size - n + 1).times { |i|
                break if words[0][n + i] != words[-1][n + i]
              }
              s = (words[0][n, i] or '')
              @out.print s
              @stroke << s
              @out.print "\r\n", words.join(' '), "\r\n"
            end
          when ?\b then
            next if @stroke.empty?
            @stroke.slice!(-1)
            @out.print "\b \b"
          else
            throw :close if c != ?\t and c < ?\s
            @stroke << c.chr
            @out.print c.chr
          end
        }
      ensure
        @out.print "(quit)\r\n" if @ready
        do_stty '-raw echo'
      end
    }
  end

end




class RbcdVerbose < Rbcd

  def initialize(file, observer, out = STDOUT)
    super file, observer
    @out = out
  end

  def disc_set(slot, disc)
    @out.print "disc found on slot \##{slot}: #{disc}\n" if $VERBOSE
    super
  end

  def disc_removed(slot, disc)
    @out.print "disc lost on slot \##{slot}: #{disc}\n" if $VERBOSE
    super
  end

  def audio_started(track)
    @out.print "audio_started: #{track.inspect}\n" if $VERBOSE
    super
  end

  def track_changed(track, prev_track)
    if $VERBOSE then
      @out.print "track_changed: #{prev_track.inspect} -> #{track.inspect}\n"
    end
    super
  end

  def audio_completed(prev_track)
    @out.print "audio_completed: #{prev_track.inspect}\n" if $VERBOSE
    super
  end

  def audio_stopped(prev_track)
    @out.print "audio_stopped: #{prev_track.inspect}\n" if $VERBOSE
    super
  end

  #def audio_playing(track, disc_addr, track_time)
  #  super
  #end

  def save_disc(*args)
    @out.print "save_disc: #{args.join(', ')})\n" if $VERBOSE
  end

end




class RbcdTTY

  def initialize(file)
    @out = Console.new
    @list = []
    @rbcd = RbcdVerbose.new(file, self, @out)
  end


  private

  def ls(range = nil)
    if @list.empty? then
      $out.print "no track.\n"
    else
      range = 0...(@list.size) unless range
      for n in range
        t = @list[n]
        break unless t
        @out.printf "%2d: %s #%d %2d %02d:%02d.%02d - %02d:%02d.%02d %s\n",
          n, (t.enable? ? " " : "X"), t.disc.slot_no, t.track_no,
          t.start_addr.min, t.start_addr.sec, t.start_addr.frame,
          t.next_addr.min, t.next_addr.sec, t.next_addr.frame,
          t.title
      end
    end
  end


  public

  def list_changed(tracks)
    @out.print "list_changed:\n" if $VERBOSE
    @list = tracks
  end


  def current_changed(track)
    @out.print "current_changed: #{track.inspect}\n" if $VERBOSE
    if track then
      @out.prompt = sprintf('#%1d %02d/%02d', track.disc.slot_no,
                            track.track_no, track.disc.max_track_no)
    elsif @list.empty? then
      @out.prompt = 'no disc'
    else
      @out.prompt = 'no current track'
    end
  end


  def now_playing(track, disc_addr, track_time)
    disc = track.disc
    @out.prompt = sprintf("#%1d %02d/%02d %02d:%02d.%02d/%02d:%02d.%02d " +
                          "(%02d:%02d.%02d/%02d:%02d.%02d)",
                          disc.slot_no, track.track_no, disc.max_track_no,
                          track_time.min, track_time.sec, track_time.frame,
                          track.size.min, track.size.sec, track.size.frame,
                          disc_addr.min, disc_addr.sec, disc_addr.frame,
                          disc.size.min, disc.size.sec, disc.size.frame)
  end


  def on_eject
    @rbcd.eject
  end

  def on_play(offset = nil)
    if offset then
      if offset[0] == ?+ then
        offset = offset[1, offset.size].to_i
        @rbcd.play_from offset
      else
        track = offset.to_i
        raise "out of range" unless @list[track]
        @list[track].play
      end
    else
      @rbcd.play
    end
  end

  def on_stop
    @rbcd.stop
  end

  def on_pause
    @rbcd.pause
  end

  def on_next
    @rbcd.next_track
  end

  def on_prev
    @rbcd.prev_track
  end

  def on_volume(volume = nil, balance = nil)
    unless volume then
      @out.printf("volume: %d  balance: %.3f\n", *@rbcd.get_volume)
    else
      balance = @rbcd.get_volume[1] unless balance
      @rbcd.set_volume volume, balance
    end
  end

  def on_repeat(f)
    @rbcd.repeat = f
    @out.print "repeat: #{@rbcd.repeat.inspect}\n"
  end

  def on_skip(track)
    if track[0] == ?- then
      track = -(track.to_i)
      raise "out of range" unless @list[track]
      @list[track].enable = true
    else
      track = track.to_i
      raise "out of range" unless @list[track]
      @list[track].enable = false
    end
    ls track..track
  end

  def on_ls(track = nil)
    unless track then
      ls
    else
      track = track.to_i
      ls track..track
    end
  end

  def on_mv(track, track_to)
    track = track.to_i
    track_to = track_to.to_i
    raise "out of range" unless @list[track] and @list[track_to]
    @rbcd.move_track @list[track], @list[track_to], true
    ls
  end

  def on_sort
    @rbcd.sort_tracks
    ls
  end

  def on_save
    @rbcd.save_all_discs
    @out.print "done.\n"
  end


  Commands = {
    'eject'  => [ :on_eject,  0, 0 ],
    'stop'   => [ :on_stop,   0, 0 ],
    'pause'  => [ :on_pause,  0, 0 ],
    'play'   => [ :on_play,   0, 1 ],
    'next'   => [ :on_next,   0, 0 ],
    'prev'   => [ :on_prev,   0, 0 ],
    'volume' => [ :on_volume, 0, 2 ],
    'repeat' => [ :on_repeat, 1, 0 ],
    'mv'     => [ :on_mv,     2, 0 ],
    'skip'   => [ :on_skip,   1, 0 ],
    'ls'     => [ :on_ls,     0, 1 ],
    'sort'   => [ :on_sort,   0, 0 ],
    'save'   => [ :on_save,   0, 0 ],
  }

  def main
    @out.add_word Commands.keys
    @out.on_idle { @rbcd.dispatch }
    @out.prompt = 'no disc'
    @out.start { |s|
      command, *argv = s.strip.split(/\s+/)
      status = Commands[command]
      if status.nil? then
        @out.print "what??\n"
      elsif argv.size < status[1] then
        @out.print "too few arguments\n"
      elsif argv.size > status[1] + status[2] then
        @out.print "too many arguments\n"
      else
        begin
          send(status[0], *argv)
        rescue Rbcd::Error, RuntimeError
          @out.print "Error: #{$!}\n"
        end
      end
    }
  end

end

RbcdTty = RbcdTTY




if $0 == __FILE__ then
  if ARGV.size > 1 then
    print "usage: #{$0} [<device>]\n"
  else
    RbcdTTY.new((ARGV[0] or CDROM::DefaultDevice)).main
  end
end
