#
# rbcd.rb - Ruby CD player
#
#   Copyright (C) Ueno Katsuhiro 2000
#
# $Id: rbcd.rb,v 1.1.1.1 2000/10/15 06:13:33 katsu Exp $
#

require 'cdrom'
require 'delegate'


class Rbcd

  PlayerName = 'rbcd'
  Version = '0.0.1'


  class Bug < Exception ;  end
  class Error < StandardError ;  end



  class Track < DelegateClass(CDROM::Track)

    def initialize(disc, track, next_addr)
      @disc = disc
      @next_addr = next_addr
      @title = ''
      @ext = ''
      @enable = true
      super track
    end

    attr_reader :disc, :enable, :title, :ext, :next_addr
    alias enable? enable

    alias ==   equal?
    alias ===  equal?

    def enable=(f)
      @enable = (f != false and f != nil)
    end

    def title=(t)
      @title = t.to_s
    end

    def ext=(t)
      @ext = t.to_s
    end

    def size
      @next_addr - start_addr
    end


    def seq?(prev_track)
      @disc == prev_track.disc and start_addr == prev_track.next_addr
    end

    include Comparable

    def <=>(other)
      if @disc != other.disc then
        @disc.slot_no <=> other.disc.slot_no
      else
        track_no <=> other.track_no
      end
    end


    def eject
      @disc = nil
    end

    def ejected?
      @disc.nil?
    end


    def play(offset = 0)
      if offset >= 0 then
        addr = start_addr + offset
        limit = @next_addr - 2
        addr = limit if addr >= limit
      else
        addr = @next_addr - (1 - offset)
        addr = start_addr if addr < start_addr
      end
      @disc.play addr
    end

  end




  class Disc

    def initialize(dev, tracks)
      @dev = dev
      raise Bug, "invalid tracks" if tracks[-1] and not tracks[-1].leadout?
      @size = tracks[-1].start_addr
      @tracks = []
      tracks.each_with_index { |t,n|
        break unless tracks[n+1]
        if t.track_no > 0 then
          @tracks[t.track_no] = Track.new(self, t, tracks[n+1].start_addr)
        end
      }
      raise Bug, "no track in this disc" if @tracks.empty?
      @title = @artist = @ext = ''
    end

    attr_reader :size, :title, :artist, :ext

    alias ==   equal?
    alias ===  equal?


    def title=(s)
      @title = s.to_s
    end

    def artist=(s)
      @artist = s.to_s
    end

    def ext=(s)
      @ext = s.to_s
    end

    def max_track_no
      @tracks.size - 1
    end


    include Enumerable

    def each
      @tracks.each { |t| yield t }
    end

    def to_a
      @tracks.dup
    end

    def [](n)
      @tracks[n]
    end

    def tracks
      @tracks.reject { |t| t.nil? }
    end

    def audio_tracks
      @tracks.reject { |t| t.nil? or not t.audio_track? }
    end


    def eject
      return unless @dev
      @tracks.each { |t| t.eject if t }
      @dev = nil
    end

    def ejected?
      @dev.nil?
    end

    def slot_no
      @dev.each { |d,n| return n if d == self }
      nil
    end


    def play(addr)
      @dev.play self, addr
    end

  end




  class CDPlayer

    # CD changer is not completely supported yet, I think.
    # Well, I don't have that ...

    def initialize(file, observer)
      @dev = CDROM.open(file)
      @observer = observer
      @discs = Array.new(@dev.num_slots)
      @current_track = nil
    end


    def each
      @discs.each_with_index { |d,n| yield d, n if d }
    end


    def dispatch
      @discs.each_with_index { |disc,n|
        if @dev.ready? n then
          unless disc then
            @dev.change_slot n
            disc = Disc.new(self, @dev.get_tracks)
            @discs[n] = disc
            @observer.disc_set n, disc
          end
        else
          if disc then
            if @current_track then
              # @dev.stop
              track = @current_track
              @current_track = nil
              @observer.audio_stopped track
            end
            disc.eject
            @discs[n] = nil
            @observer.disc_removed n, disc
          end
        end
      }
      disc = @discs[@dev.current_slot]
      return unless disc

      loop {
        status = @dev.get_status
        if status.paused? then
          # nothing to do
        elsif status.playing? then
          track = disc[status.track_no]
          if track.nil? or not track.audio_track? then
            raise Bug, "Hmm... are you playing a data track?"
          elsif track != @current_track then
            prev_track = @current_track
            @current_track = track
            if prev_track then
              @observer.track_changed track, prev_track
            else
              @observer.audio_started track
            end
            next
          end
          if @current_track then
            @observer.audio_playing track, status.addr, status.track_time
          end
        elsif @current_track then
          prev_track = @current_track
          @current_track = nil
          if status.completed? then
            @observer.audio_completed prev_track
          else
            @observer.audio_stopped prev_track
          end
        end
        break
      }
    end


    def eject(slot = nil)
      @dev.change_slot slot
      begin
        @dev.eject
      rescue
      end
    end


    def stop
      @dev.stop_audio
    end


    def pause
      @dev.pause_audio
    end


    def resume
      @dev.resume_audio
    end


    def play(disc, addr)
      n = @discs.index(disc)
      raise Bug, "no such disc in #{self}: #{disc}" unless n
      @dev.change_slot n
      @dev.play_audio addr
    end


    def get_status
      @dev.get_status
    end


    def set_volume(volume, balance)
      raise Error, "volume out of range" if volume < 0 or volume > 255
      raise Error, "balance out of range" if balance < 0 or balance > 100
      n = ((balance - 50) * volume) / 50
      if n < 0 then
        l, r = volume, volume + n
      else
        l, r = volume - n, volume
      end
      @dev.set_volume l.to_i, r.to_i
    end


    def get_volume
      vols = @dev.get_volume
      volume = vols.max
      if volume == 0 then
        balance = 50.0
      else
        balance = Float(vols[1] - vols[0]) * 50 / volume + 50
      end
      [ volume, balance ]
    end

  end




  class CDDB

    DBDir = ENV['HOME'] + File::Separator + '.rbcd' + File::Separator


    #def initialize
    #end


    def disc_id(disc)
      a = disc.tracks
      n, t = 0, 0
      a.each_with_index { |track,i|
        addr = track.start_addr
        if a[i+1] then
          next_addr = a[i+1].start_addr
        else
          next_addr = disc.size
        end
        addr.real.seconds.to_i.to_s.each_byte { |c| n += c - ?0 }
        t += next_addr.seconds.to_i - track.addr.seconds.to_i
      }
      '%08x' % ((n % 0xff) << 24 | t << 8 | (disc.max_track_no))
    end


    def load_disc(disc)
      disc_id = disc_id(disc)
      filename = DBDir + disc_id
      begin
        f = File.open(filename)
      rescue Errno::ENOENT
        return disc.audio_tracks
      end
      playlist = nil
      f.each { |line|
        case line
        when /\ADISCID=/ then
          raise 'invalid DISCID' unless $'.strip == disc_id
        when /\ADTITLE=/ then
          title, artist = $'.strip.split(/\s*\/\s*(?=[^\/]*\z)/, 2)
          artist = '' unless artist
          disc.title = title
          disc.artist = artist
        when /\AEXTD=/ then
          disc.ext = $'.strip
        when /\ATTITLE(\d+)=/ then
          track = disc[$1.to_i + 1]
          track.title = $'.strip if track
        when /\AEXTT(\d+)=/ then
          track = disc[$1.to_i + 1]
          track.ext = $'.strip if track
        when /\ASKIP=/ then
          $'.strip.split(/\s+/).each { |i|
            track = disc[i.to_i + 1]
            track.enable = false if track
          }
        when /\APLAYLIST=/, /\APLAYORDER=/ then
          s = $'.strip
          unless s.empty? then
            a = disc.to_a
            playlist = s.split(/\s+/).collect { |i|
              i = i.to_i + 1
              t = a[i]
              a[i] = nil
              t = nil if t and not t.audio_track?
              t
            }
            a.compact!
            playlist.compact!
            playlist.concat a
          end
        else
          # comments or unknown properties. just ignore them.
        end
      }
      playlist or disc.audio_tracks
    rescue
      STDERR.printf "#{$0}: #{filename}: #{$!}\n" if $VERBOSE
    end


    def save_disc(disc, order, force = false)
      audio = disc.audio_tracks
      disc_id = disc_id(disc)
      return if not force and (audio.empty? or disc.title.empty?)
      filename = DBDir + disc_id
      begin
        f = File.open(filename, 'w')
      rescue Errno::ENOENT
        Dir.mkdir DBDir
        f = File.open(filename, 'w')
      end
      tracks = Array.new(disc.max_track_no)
      audio.each { |t| tracks[t.track_no - 1] = t if t.track_no >= 1 }
      f.print("# xmcd CD Database Entry\n" +
              "#\n" +
              "# Track frame offsets:" +
              disc.collect{ |t| t and t.start_addr.to_i }.join("\n# "), "\n" +
              "#\n" +
              "# Disc length: ", disc.size.seconds.to_i, " seconds\n" +
              "#\n" +
              "# Revision: 0\n" +
              "# Submitted via: ", PlayerName, ' ', Version, "\n" +
              "#\n")
      f.print 'DISCID=', disc_id, "\n"
      title = disc.title
      title += ' / ' + disc.artist unless disc.artist.empty?
      f.print 'DTITLE=', title, "\n"
      tracks.each_with_index { |trk,n|
        f.print 'TTITLE', n, '=', (trk and trk.title).to_s, "\n"
      }
      f.print 'EXTD=', disc.ext, "\n"
      tracks.each_with_index { |trk,n|
        f.print 'TEXT', n, '=', (trk and trk.ext).to_s, "\n"
      }
      f.print "PLAYORDER="
      f.print order.collect{ |n| n-1 }.join(' ') unless order == order.sort
      f.print "\n"
      if audio.find { |t| not t.enable? } then
        f.print "SKIP=", audio.reject{ |t| t.enable?
        }.collect{ |t| t.track_no - 1 }.join(' '), "\n"
      end
      f.close
    rescue
      STDERR.printf "#{$0}: #{filename}: #{$!}\n" if $VERBOSE
    end

  end




  class TrackList

    def initialize
      @tracks = []
      @current = nil
    end


    include Enumerable

    def each
      @tracks.each { |t| yield t }
    end

    def [](n)
      @tracks[n]
    end

    def empty?
      @tracks.empty?
    end

    def size
      @tracks.size
    end

    def to_a
      @tracks.dup
    end


    private

    def first_enable_track(offset = 0)
      if offset == 0 then
        @tracks.collect{ |t| t.enable? }.index(true)
      else
        n = @tracks[offset, @tracks.size].collect{ |t| t.enable? }.index(true)
        n += offset if n
        n
      end
    end


    def last_enable_track(limit = @tracks.size)
      if limit == @tracks.size then
        @tracks.collect{ |t| t.enable? }.rindex(true)
      else
        @tracks[0, limit].collect{ |t| t.enable? }.rindex(true)
      end
    end


    public

    def append(tracks)
      @tracks.concat tracks
      if @current then
        nil
      else
        @current = first_enable_track
      end
    end


    def eject!
      unless @current then
        @tracks.reject! { |t| t.ejected? }
        nil
      else
        pre = @tracks[0, @current].reject { |t| t.ejected? }
        post = @tracks[@current, @tracks.size].reject { |t| t.ejected? }
        @tracks = pre + post
        cur = first_enable_track(pre.size)
        cur = last_enable_track(pre.size) unless cur
        if cur == @current then
          nil
        else
          @current = cur
        end
      end
    end


    def move_track(track, track_to)
      return if track == track_to
      trk = current_track
      @tracks.delete(track) { raise Bug, "no such track in #{self}: #{track}" }
      unless track_to then
        @tracks.push track
      else
        n = @tracks.index(track_to)
        raise Bug, "no such track in #{self}: #{track_to}" unless n
        @tracks[n, 0] = track
      end
      if trk == track then
        @current = n
      else
        nil
      end
    end


    def sort!
      trk = current_track
      @tracks.sort!
      @current = @tracks.index(trk)
      nil
    end


    def select_track(track)
      unless track then
        @current = first_enable_track
        track = (@current and @tracks[@current])
      else
        @current = @tracks.index(track)
        raise Bug, "no such track in #{self}: #{track}" unless @current
      end
      track
    end


    def current_track
      @current and @tracks[@current]
    end


    def first_track
      n = first_enable_track
      n and @tracks[n]
    end

    def last_track
      n = last_enable_track
      n and @tracks[n]
    end

    def next_track
      if @current then
        n = first_enable_track(@current + 1)
        n and @tracks[n]
      else
        nil
      end
    end

    def prev_track
      if @current then
        n = last_enable_track(@current)
        n and @tracks[n]
      else
        nil
      end
    end

  end




  def initialize(file, observer)
    @observer = observer
    @tracklist = TrackList.new
    @cddb = CDDB.new
    @repeat = false
    @dev = CDPlayer.new(file, self)
  end

  attr_reader :repeat

  def repeat=(f)
    @repeat = (f != false and f != nil)
  end


  def dispatch
    @dev.dispatch
  end


  private

  def list_changed
    @observer.list_changed @tracklist.to_a
  end

  def current_changed
    @observer.current_changed @tracklist.current_track
  end


  public

  def disc_set(slot, disc)
    tracks = @cddb.load_disc(disc)
    e = @tracklist.empty?
    f = @tracklist.append(tracks)
    list_changed
    current_changed if e or f
  end


  def disc_removed(slot, disc)
    f = @tracklist.eject!
    e = @tracklist.empty?
    list_changed
    current_changed if f or e
  end


  def audio_started(track)
    @tracklist.select_track track
    current_changed
  end


  def track_changed(track, prev_track)
    unless track.seq? prev_track then
      @tracklist.select_track track
      current_changed
    else
      next_track = @tracklist.next_track
      if next_track.nil? and @repeat then
        next_track = (@tracklist.first_track or prev_track)
      end
      if next_track.nil? then
        @dev.stop
      elsif track != next_track then
        next_track.play
      else
        @tracklist.select_track next_track
        current_changed
      end
    end
  end


  def audio_completed(prev_track)
    next_track = @tracklist.first_track
    if @repeat and next_track then
      next_track.play
    elsif @repeat then
      prev_track.play
    else
      @tracklist.select_track nil
      current_changed
    end
  end


  def audio_stopped(prev_track)
    @tracklist.select_track nil
    current_changed
  end


  def audio_playing(track, disc_addr, track_time)
    @observer.now_playing track, disc_addr, track_time
  end


  public

  def eject(slot = nil)
    @dev.eject slot
  end


  def stop
    @dev.stop
  end


  def pause
    case @dev.get_status.status
    when 'playing' then
      @dev.pause
    when 'paused' then
      @dev.resume
    end
  end


  def play
    case @dev.get_status.status
    when 'playing' then
      # nothing to do
    when 'paused' then
      @dev.resume
    else
      track = (@tracklist.current_track or @tracklist.first_track)
      track.play if track
    end
  end


  def play_from(offset)
    track = @tracklist.current_track
    return unless track
    track.play offset
  end


  def next_track
    track = @tracklist.next_track
    track = @tracklist.first_track unless track
    return unless track
    status = @dev.get_status
    if status.playing? or status.paused? then
      track.play
    else
      @tracklist.select_track track
      current_changed
    end
  end


  def prev_track
    track = @tracklist.prev_track
    track = @tracklist.last_track unless track
    return unless track
    status = @dev.get_status
    if status.playing? or status.paused? then
      track.play
    else
      @tracklist.select_track track
      current_changed
    end
  end


  def move_track(track, track_to, notify = false)
    @tracklist.move_track track, track_to
    list_changed if notify
  end


  def sort_tracks(notify = true)
    @tracklist.sort!
    list_changed if notify
  end


  def disc_id(disc)
    @cddb.disc_id disc
  end


  def save_disc(disc, force = false)
    order = @tracklist.reject{ |t| t.disc != disc }.collect{ |t| t.track_no }
    @cddb.save_disc disc, order, force
  end


  def save_all_discs(force = false)
    @dev.each { |disc,n| save_disc disc, force }
  end


  def set_volume(volume, balance)
    @dev.set_volume volume, balance
  end


  def get_volume
    @dev.get_volume
  end

end
